# External Entities

This module allows you to connect to and use datasets from external sources in
your Drupal website. While content lives externally, Drupal sees those remote
entities as internal. This makes it possible to alter the field displays, add
references, comments, path aliases, share buttons, web-shop products and more.

For a full description of the module, visit the 
[project page](https://www.drupal.org/project/external_entities).

Submit bug reports and feature suggestions, or track changes in the
[issue queue](https://www.drupal.org/project/issues/external_entities).


## Table of contents

- Requirements
- Installation
- Configuration
- How it works
- Recommended modules
- v2 to v3 changes
- Maintainers


## Requirements

This module requires no modules outside of Drupal core but will require the
JsonPath PHP library ([galbar/jsonpath](https://github.com/Galbar/JsonPath-PHP))
which is automatically installed by composer.


## Installation

Install as you would normally install a contributed Drupal module. For further
information, see
[Installing Drupal Modules](https://www.drupal.org/docs/extending-drupal/installing-drupal-modules).


## Configuration

1. Enable the module at Administration > Extend.
2. To enable integration of external entities, create a new external entity on
   admin/structure/external-entity-types.
3. Configure the external source, add (Drupal) fields, save and edit the
   external entity again to configure field mappings.

For more information, see the documentation page about
[External Entities](https://www.drupal.org/docs/contributed-modules/external-entities).


## How it works

### Saving

- non-mapped source field values are left "as they are".
- Drupal fields mapped to constant values are not saved.
- Multiple values for a given source field are replaced by the Drupal values.
- Mapped Drupal fields that have empty values replace/clear source values.
- Drupal fields mapped to non-existing source fields are saved (if the source
  storage takes them into account).

## Recommended modules

- [Search API](https://www.drupal.org/project/search_api), when enabled
it is possible to make overviews of external entities.

- [External Entity Manager](https://www.drupal.org/project/xnttmanager), which
allows to inspect available external fields and their mapping, to synchronize
local data with external data and auto-generate associated annotation entities.

- [External Entity Files storage](https://www.drupal.org/project/xnttfiles), a
file storage client plugin which also allows to map Drupal File and Image fields
to external URIs (URLs or stream paths).

- [External Entities Database storage](https://www.drupal.org/project/xnttdb),
an external database storage client plugin which allows use external databases
(both mySQL and Postgres, regardless what Drupal is using) as external data
sources.

- [External Entities Views](https://www.drupal.org/project/xnttviews),
an experimental plugin that enables the use of external entities with Views.

Find more contributed modules at the
[External Entities plugins page](https://www.drupal.org/docs/contributed-modules/external-entities/external-entities-plugins)


## v2 to v3 changes

v3 has been introduced to allow major External Entities upgrades (through
breaking changes) while keeping v2 to let previous system continue working
without being required to upgrade. The v3 changes come from raised issues,
improvements and new feature requests, after several years of use of External
Entities in various contexts.

List of *BREAKING CHANGES* between External Entities v2 and v3:
- Global changes:
  - Type hinting has been added to v3 wherever possible.
  - Storage class name simplification: "ExternalEntityStorageClient*" become
    "StorageClient*".
  - Storage client plugin type name changed from "ExternalEntityStorageClient"
    to "StorageClient".
- Field Mapper
  - Removed unnecessary abstraction layers for field mappers:
    - "ExpressionFieldMapperInterface" has been merged to
      "FieldMapperInterface".
    - "ConfigurableExpressionFieldMapperBase" and "ExpressionFieldMapperBase"
      have beend merge to "FieldMapperBase".
    Those layers came from a discussion when introducing field mapper plugins:
    https://www.drupal.org/project/external_entities/issues/3079012#comment-13437754
    Field mapper plugins should now use "FieldMapperBase" class as base.
    The rational behind this change is that those layers were adding complexity
    to the code since some methods of "ExpressionFieldMapperInterface" were
    required at some places where "FieldMapperInterface" was used. It would be
    hard to choose where to use one interface or the other while only one is
    used by plugins and current external entities implementation forces the use
    of plugins for field mappers. If one day, someone would create external
    entities type from code and want to force the use of a non-plugin field
    mapper, either this person would have to turn it into a plugin (which would
    not be so complicated) or raise an issue with explanations on the use case.
  - "createRawDataFromEntityValues" has an additional parameter
    $original_values that should be the source array which will be overriden
    by the new values.
  - "getFieldPropertyMapping" has been removed and can be replaced by
    "getFieldMapping" as the possible return types would be the same now.
  - "getConstantMappingPrefix" has been removed as the mapping type (ie. for
    instance FieldMapperInterface::MAPPING_CONSTANT) is now provided with the
    mapping string when using "getFieldMapping".
  - "extractFieldValuesFromRawData" and "extractFieldPropertyValuesFromRawData"
    are now part of the interface and made public but are using the string field
    name as first argument instead of the field definition.
  - "getMappableFields" stores mappable fields in class member "mappableFields".
- REST client:
  - Method getListQueryParameters() does not have "$start" and "$length"
    parameters. It does not handle paging anymore as well: paging is managed by
    a modified getPagingQueryParameters() method.
  - Method getPagingQueryParameters() does not return an array of paging
    parameters but an array of paging parameters for multiple queries to map
    client paging to server paging (with different page sizes).
  - The REST client does not save data using "application/x-www-form-urlencoded"
    data by default anymore. It now uses the REST endpoint encoding type set by
    the "Endpoint data encoding format" setting.

List of v3 new features:
- Global:
  - New method to set and retrieve source raw data from an external entity:
    ´setSourceRawData(array)´ and ´getSourceRawData()´. When external entities
    are saved, source fields that were not mapped to Drupal fields are kept.
  - Plugins (Storage Clients and Field Mappers) can now rely on the method
    submitConfigurationForm() rather than validateConfigurationForm() to set
    their configuration changes while still use validateConfigurationForm() to
    validate/invalidate form submission.
  - Added support for debugging external entities (loading, mapping, saving).
- REST client:
  - New endpoint options to support specific URLs to fetch a single entity or to
    count entities.
  - Support for any response structure using JSONPath to locate data in result
    sets.
  - Improved pager settings to support server limits and paging starting from 1.
  - Better support for authentication. Bearer token built-in support. API key
    support through query string. Support for custom authentication header.
  - Added support for custom headers. Ex.: "Accept: application/vnd.api+json".
  - Additional parameters can now be sent either as a query string or as POST
    data using the endpoint data encoding format. It is possible to mix POST and
    GET parameters using a question mark to specify query parameters in the
    endpoint URL.
  - Added an entity identifier placeholder "{id}" replaced in single entity
    endpoint URL and parameters.

### Upgrading v2 storage clients

- Replace ResponseDecoderFactoryInterface by EntityFieldManagerInterface in
  both __constructor() and create() methods unless you use
  ResponseDecoderFactoryInterface. In that later case, you will need to do the
  replacement in the __constructor() when calling parent class constructor and
  add EntityFieldManagerInterface to your __constructor() and create() methods.
  You can get an entity field manager from core services using:
  ```
  $container->get('entity_field.manager')
  ```
  Don't forget to update code documentation and to add the import:
  ```
  use Drupal\Core\Entity\EntityFieldManagerInterface;
  ```
  For the explanations, ResponseDecoderFactoryInterface was only needed by REST
  clients and was removed from the StorageClientInterface. On the other side,
  EntityFieldManagerInterface was needed by the new features added to
  EntityFieldManagerInterface for query filtering.
- Rename query() and countQuery() into querySource() and countQuerySource()
  respectively. This should be sufficient unless you already tried to translate
  Drupal field names into mapped source field names in your code logic. In that
  case, you may need to do some more code adjustments.
- Implement transliterateDrupalFilters() method. To start, you can use the
  basic but unefficient implementation:
  ```
  public function transliterateDrupalFilters(array $parameters) :array {
    // @todo: To really implement for better performances.
    return ['source' => [], 'drupal' => $parameters];
  }
  ```
  When implementing, you may find helpfull to use
  $this->externalEntityType->getFieldMapper() and $this->entityFieldManager new
  class member as well as the getFieldDefinitions() and getFieldMappings() new
  methods. The storage client "knows" which operators are supported by its
  storage source (list of used operators: see StorageClientInterface::query()
  documentation and Drupal Query API documentation:
  https://www.drupal.org/docs/8/api/database-api/dynamic-queries/conditions and
  https://api.drupal.org/api/drupal/core%21lib%21Drupal%21Core%21Entity%21Query%21QueryInterface.php/function/QueryInterface%3A%3Acondition/10).
  Then, some operators may or may not be supported by some source fields. While
  the $parameters parameter containes filter on *Drupal* field names, it will be
  required to match those to source fields using the field mapper
  ($this->externalEntityType->getFieldMapper() and getFieldMappings() method).
  And to tell if a given source field can support a given operator, you may rely
  on the Drupal mapped field type given by the user to guess the source field
  type ($this->entityFieldManager and getFieldDefinitions() method).

### Upgrading v2 field mappers

To be explained...

## Maintainers

Current maintainers:
- Raf Philtjens - [rp7](https://www.drupal.org/u/rp7)
- Clemens Tolboom - [clemens.tolboom](https://www.drupal.org/u/clemenstolboom)
- Hanno Lans - [Hanno](https://www.drupal.org/u/hanno)
- Rodrigo Aguilera - [rodrigoaguilera](https://www.drupal.org/u/rodrigoaguilera)
- Jelle Sebreghts - [Jelle_S](https://www.drupal.org/u/jelle_s)
- Patrick van Efferen - [pefferen](https://www.drupal.org/u/pefferen)
- Peter Grond - [pgrond](https://www.drupal.org/u/pgrond)
- Valentin Guignon - [guignonv](https://www.drupal.org/u/guignonv)

This project has been sponsored by:
- Attiks (Peter Droogmans) is no nonsense company with a highly skilled
  technical and graphical staff. We handle both on- and offline communication,
  web development and print design are our middle names. We work
  internationally and are based in Antwerp, Belgium.
- Datascape (Hanno Lans) is involved in Drupal projects with an advanced
  information architecture, with accessibility, open content and open source in
  mind.
- Kunstmuseum The Hague has a leading collection of modern and contemporary art,
  fashion and decorative arts. It is also the international 
  home of Piet Mondrian,
  with no fewer than 300 works by the famous Dutch artist in its collection.
-The Alliance Bioversity - CIAT (guignonv), a CGIAR center.
