<?php

namespace Drupal\Tests\external_entities\Functional;

use Drupal\external_entities\FieldMapper\FieldMapperInterface;
use Drupal\filter\Entity\FilterFormat;

/**
 * Tests Rest storage client for external entity.
 *
 * @group ExternalEntities
 */
class RestClientFunctionalTest extends ExternalEntitiesBrowserTestBase {

  /**
   * A user with administration permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $account;

  /**
   * The entity storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $storage;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    global $base_url;
    $this->storage = $this->container->get('entity_type.manager')->getStorage('external_entity_type');

    // Setup datasets.
    $xntt_json_controller = $this
      ->container
      ->get('controller_resolver')
      ->getControllerFromDefinition(
        '\Drupal\external_entities_test\Controller\ExternalEntitiesJsonController::listItems'
      )[0];
    $xntt_json_controller->setRawData('simple', [
      '2596b1ba-43bb-4440-9f0c-f1974f733336' => [
        'uuid' => '2596b1ba-43bb-4440-9f0c-f1974f733336',
        'title' => 'Simple title 1',
        'short_text' => 'Just a short string',
        'rich_text' => '<h2>Some HTML tags</h2>',
        'rich_text_2' => '<h2>Other HTML tags</h2>',
        'status' => TRUE,
        'refs' => [
          'ref2',
          'ref3',
        ],
      ],
      '2596b1ba-43bb-4440-9f0c-f1974f733337' => [
        'uuid' => '2596b1ba-43bb-4440-9f0c-f1974f733337',
        'title' => 'Simple title 2',
        'short_text' => 'Just another short string',
        'status' => FALSE,
      ],
    ]);
    $xntt_json_controller->setRawData('ref', [
      'ref1' => [
        'uuid' => 'ref1',
        'label' => 'Term 1',
      ],
      'ref2' => [
        'uuid' => 'ref2',
        'label' => 'Term 2',
      ],
      'ref3' => [
        'uuid' => 'ref3',
        'label' => 'Term 3',
      ],
      'ref4' => [
        'uuid' => 'ref4',
        'label' => 'Term 4',
      ],
    ]);

    /** @var \Drupal\external_entities\Entity\ExternalEntityType $ref */
    $ref = $this->container->get('entity_type.manager')->getStorage('external_entity_type')->create([
      'id' => 'ref',
      'label' => 'Ref',
      'label_plural' => 'Refs',
      'description' => '',
      'read_only' => FALSE,
      'generate_aliases' => FALSE,
      'persistent_cache_max_age' => 0,
    ]);

    // Should set client 0 by default.
    $ref->setStorageClientId('rest');
    $ref->setStorageClientConfig([
      'endpoint' => $base_url . '/external-entities-test/ref',
      'endpoint_options' => [
        'single' => '',
        'count' => '',
        'count_mode' => NULL,
        'cache' => FALSE,
        'limit_qcount' => 0,
        'limit_qtime' => 0,
      ],
      'response_format' => 'json',
      'data_path' => [
        'list' => '',
        'single' => '',
        'keyed_by_id' => FALSE,
        'count' => '',
      ],
      'pager' => [
        'default_limit' => 50,
        'page_parameter' => '',
        'page_parameter_type' => 'pagenum',
        'page_start_one' => FALSE,
        'page_size_parameter' => '',
        'page_size_parameter_type' => 'pagesize',
      ],
      'api_key' => [
        'type' => 'none',
        'header_name' => '',
        'key' => '',
      ],
      'http' => [
        'headers' => '',
      ],
      'parameters' => [
        'list' => [],
        'list_param_mode' => 'query',
        'single' => [],
        'single_param_mode' => 'query',
      ],
    ]);

    $ref->setFieldMapperId('simple');
    $ref->setFieldMapperConfig([
      'field_mappings' => [
        'id' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'uuid' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'title' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'label',
          ],
        ],
      ],
    ]);
    $ref->save();
    $full_html_format = FilterFormat::create([
      'format' => 'full_html',
      'name' => 'Full HTML',
      'weight' => 1,
      'filters' => [],
    ]);
    $full_html_format->save();
    /** @var \Drupal\external_entities\Entity\ExternalEntityType $type */
    $type = $this->container->get('entity_type.manager')->getStorage('external_entity_type')->create([
      'id' => 'simple_external_entity',
      'label' => 'Simple external entity',
      'label_plural' => 'Simple external entities',
      'description' => '',
      'read_only' => FALSE,
      'generate_aliases' => FALSE,
      'persistent_cache_max_age' => 0,
    ]);

    // Explicitely sets client 0.
    $type->setStorageClientId('rest', 0);
    $type->setStorageClientConfig([
      'endpoint' => $base_url . '/external-entities-test/simple',
      'endpoint_options' => [
        'single' => '',
        'count' => '',
        'count_mode' => NULL,
        'cache' => FALSE,
        'limit_qcount' => 0,
        'limit_qtime' => 0,
      ],
      'response_format' => 'json',
      'data_path' => [
        'list' => '',
        'single' => '',
        'keyed_by_id' => FALSE,
        'count' => '',
      ],
      'pager' => [
        'default_limit' => 50,
        'page_parameter' => '',
        'page_parameter_type' => 'pagenum',
        'page_start_one' => FALSE,
        'page_size_parameter' => '',
        'page_size_parameter_type' => 'pagesize',
      ],
      'api_key' => [
        'type' => 'none',
        'header_name' => '',
        'key' => '',
      ],
      'http' => [
        'headers' => '',
      ],
      'parameters' => [
        'list' => [],
        'list_param_mode' => 'query',
        'single' => [],
        'single_param_mode' => 'query',
      ],
    ],
    0);

    $type->setFieldMapperId('simple');
    $type->save();

    // Add fields.
    $this->createField('simple_external_entity', 'plain_text', 'string');
    $this->createField('simple_external_entity', 'fixed_string', 'string');
    $this->createField('simple_external_entity', 'a_boolean', 'boolean');
    $this->createField('simple_external_entity', 'a_rich_text', 'text');
    $this->createField('simple_external_entity', 'a_plain_text', 'text');
    $this->createReferenceField('simple_external_entity', 'ref', 'ref', NULL, TRUE);

    $type->setFieldMapperConfig([
      'field_mappings' => [
        'id' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'uuid' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'title' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'title',
          ],
        ],
        'plain_text' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'short_text',
          ],
        ],
        'fixed_string' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_CONSTANT,
            'mapping' => 'A fixed string',
          ],
        ],
        'a_rich_text' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'rich_text',
          ],
          'format' => [
            'type' => FieldMapperInterface::MAPPING_CONSTANT,
            'mapping' => 'full_html',
          ],
        ],
        'a_plain_text' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'rich_text_2',
          ],
          'format' => [
            'type' => FieldMapperInterface::MAPPING_CONSTANT,
            'mapping' => 'plain_text',
          ],
        ],
        'a_boolean' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'status',
          ],
        ],
        'ref' => [
          'target_id' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'refs/*',
          ],
        ],
      ],
    ]);
    $type->save();

    // Create the user with all needed permissions.
    $this->account = $this->drupalCreateUser([
      'administer external entity types',
      'view simple_external_entity external entity',
      'view simple_external_entity external entity collection',
    ]);
    $this->drupalLogin($this->account);
  }

  /**
   * Tests REST storage client global features.
   *
   * Other technical features are tested into unit tests.
   */
  public function testRestClient() {
    global $base_url;
    /** @var \Drupal\Tests\WebAssert $assert */
    $assert = $this->assertSession();

    // Edit external entity.
    $this->drupalGet('admin/structure/external-entity-types');
    $this->drupalGet('admin/structure/external-entity-types/simple_external_entity');
    $assert->pageTextContains('Simple external entity');

    // @todo Check if static cache works properly: mock endpoint to make sure we
    // only got on query for the list and none for single entity loading when
    // listing and loading. And then, make sure we still load if entity is not
    // in static cache.
    $this->fillField('storages[0][storage_client_config][endpoint_options][cache]', TRUE);

    // @todo Check no URL for a single entity.
    $this->fillField('storages[0][storage_client_config][endpoint_options][single]', '');
    // @todo Check URL for a single entity with id replacement.
    $this->fillField('storages[0][storage_client_config][endpoint_options][single]', $base_url . '/external-entities-test/simple/{id}/details');
    // @todo Check URL for a single entity without {id} placeholders and no replacement (id in the query part).
    $this->fillField('storages[0][storage_client_config][endpoint_options][single]', $base_url . '/external-entities-test/simple/');

    // @todo Check other response format: serialized and yml.
    // Field storages[0][storage_client_config][response_format].
    $this->fillField('storages[0][storage_client_config][response_format]', 'yml');
    $this->fillField('storages[0][storage_client_config][response_format]', 'serialized');
    $this->fillField('storages[0][storage_client_config][response_format]', 'json');

    // @todo Test if list parameter string is properly parsed and turned into an
    // array in config.
    $this->fillField('storages[0][storage_client_config][parameters][list]', "testpropa|a_value\ntestpropb|2,3,4");

    // @todo Test if single parameter string is properly parsed and turned into
    // an array in config with '{id}' placeholders replaced at runtime.
    $this->fillField('storages[0][storage_client_config][parameters][single]', "propc|other_value\nidfield|ent{id}x");

    // @todo Test single entities keyed by their ID in a list.
    $this->fillField('storages[0][storage_client_config][data_path][keyed_by_id]', TRUE);

    // Save changes.
    $this->pressButton('edit-actions-submit');

    // @todo Implement tests (see above).
    $this->markTestIncomplete('A couple of elements remains to be tested.');
  }

}
