<?php

namespace Drupal\Tests\external_entities\Functional;

use Behat\Mink\Exception\ElementNotFoundException;
use Drupal\Component\Serialization\Json;
use Drupal\external_entities\FieldMapper\FieldMapperInterface;
use Drupal\filter\Entity\FilterFormat;

/**
 * Tests creation of a simple external entity.
 *
 * @group ExternalEntities
 */
class SimpleExternalEntityTest extends ExternalEntitiesBrowserTestBase {

  /**
   * A user with administration permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $account;

  /**
   * The entity storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $storage;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    global $base_url;
    $this->storage = $this->container->get('entity_type.manager')->getStorage('external_entity_type');

    // Setup datasets.
    $xntt_json_controller = $this
      ->container
      ->get('controller_resolver')
      ->getControllerFromDefinition(
        '\Drupal\external_entities_test\Controller\ExternalEntitiesJsonController::listItems'
      )[0];
    $xntt_json_controller->setRawData('simple', [
      '2596b1ba-43bb-4440-9f0c-f1974f733336' => [
        'uuid' => '2596b1ba-43bb-4440-9f0c-f1974f733336',
        'title' => 'Simple title 1',
        'short_text' => 'Just a short string',
        'rich_text' => '<h2>Some HTML tags</h2>',
        'rich_text_2' => '<h2>Other HTML tags</h2>',
        'status' => TRUE,
        'refs' => [
          'ref2',
          'ref3',
        ],
      ],
      '2596b1ba-43bb-4440-9f0c-f1974f733337' => [
        'uuid' => '2596b1ba-43bb-4440-9f0c-f1974f733337',
        'title' => 'Simple title 2',
        'short_text' => 'Just another short string',
        'status' => FALSE,
      ],
    ]);
    $xntt_json_controller->setRawData('ref', [
      'ref1' => [
        'uuid' => 'ref1',
        'label' => 'Term 1',
      ],
      'ref2' => [
        'uuid' => 'ref2',
        'label' => 'Term 2',
      ],
      'ref3' => [
        'uuid' => 'ref3',
        'label' => 'Term 3',
      ],
      'ref4' => [
        'uuid' => 'ref4',
        'label' => 'Term 4',
      ],
    ]);

    /** @var \Drupal\external_entities\Entity\ExternalEntityType $ref */
    $ref = $this->container->get('entity_type.manager')->getStorage('external_entity_type')->create([
      'id' => 'ref',
      'label' => 'Ref',
      'label_plural' => 'Refs',
      'description' => '',
      'read_only' => FALSE,
      'generate_aliases' => FALSE,
      'persistent_cache_max_age' => 0,
    ]);

    // Should set client 0 by default.
    $ref->setStorageClientId('rest');
    $ref->setStorageClientConfig([
      'endpoint' => $base_url . '/external-entities-test/ref',
      'endpoint_options' => [
        'single' => '',
        'count' => '',
        'count_mode' => NULL,
        'cache' => FALSE,
        'limit_qcount' => 0,
        'limit_qtime' => 0,
      ],
      'response_format' => 'json',
      'data_path' => [
        'list' => '',
        'single' => '',
        'keyed_by_id' => FALSE,
        'count' => '',
      ],
      'pager' => [
        'default_limit' => 50,
        'page_parameter' => '',
        'page_parameter_type' => 'pagenum',
        'page_start_one' => FALSE,
        'page_size_parameter' => '',
        'page_size_parameter_type' => 'pagesize',
      ],
      'api_key' => [
        'type' => 'none',
        'header_name' => '',
        'key' => '',
      ],
      'http' => [
        'headers' => '',
      ],
      'parameters' => [
        'list' => [],
        'list_param_mode' => 'query',
        'single' => [],
        'single_param_mode' => 'query',
      ],
    ]);

    $ref->setFieldMapperId('simple');
    $ref->setFieldMapperConfig([
      'field_mappings' => [
        'id' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'uuid' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'title' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'label',
          ],
        ],
      ],
    ]);
    $ref->save();
    $full_html_format = FilterFormat::create([
      'format' => 'full_html',
      'name' => 'Full HTML',
      'weight' => 1,
      'filters' => [],
    ]);
    $full_html_format->save();
    /** @var \Drupal\external_entities\Entity\ExternalEntityType $type */
    $type = $this->container->get('entity_type.manager')->getStorage('external_entity_type')->create([
      'id' => 'simple_external_entity',
      'label' => 'Simple external entity',
      'label_plural' => 'Simple external entities',
      'description' => '',
      'read_only' => FALSE,
      'generate_aliases' => FALSE,
      'persistent_cache_max_age' => 0,
    ]);

    // Explicitely sets client 0.
    $type->setStorageClientId('rest', 0);
    $type->setStorageClientConfig([
      'endpoint' => $base_url . '/external-entities-test/simple',
      'endpoint_options' => [
        'single' => '',
        'count' => '',
        'count_mode' => NULL,
        'cache' => FALSE,
        'limit_qcount' => 0,
        'limit_qtime' => 0,
      ],
      'response_format' => 'json',
      'data_path' => [
        'list' => '',
        'single' => '',
        'keyed_by_id' => FALSE,
        'count' => '',
      ],
      'pager' => [
        'default_limit' => 50,
        'page_parameter' => '',
        'page_parameter_type' => 'pagenum',
        'page_start_one' => FALSE,
        'page_size_parameter' => '',
        'page_size_parameter_type' => 'pagesize',
      ],
      'api_key' => [
        'type' => 'none',
        'header_name' => '',
        'key' => '',
      ],
      'http' => [
        'headers' => '',
      ],
      'parameters' => [
        'list' => [],
        'list_param_mode' => 'query',
        'single' => [],
        'single_param_mode' => 'query',
      ],
    ],
    0);

    $type->setFieldMapperId('simple');
    $type->save();

    // Add fields.
    $this->createField('simple_external_entity', 'plain_text', 'string');
    $this->createField('simple_external_entity', 'fixed_string', 'string');
    $this->createField('simple_external_entity', 'a_boolean', 'boolean');
    $this->createField('simple_external_entity', 'a_rich_text', 'text');
    $this->createField('simple_external_entity', 'a_plain_text', 'text');
    $this->createReferenceField('simple_external_entity', 'ref', 'ref', NULL, TRUE);

    $type->setFieldMapperConfig([
      'field_mappings' => [
        'id' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'uuid' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'uuid',
          ],
        ],
        'title' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_DIRECT,
            'mapping' => 'title',
          ],
        ],
        'plain_text' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'short_text',
          ],
        ],
        'fixed_string' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_CONSTANT,
            'mapping' => 'A fixed string',
          ],
        ],
        'a_rich_text' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'rich_text',
          ],
          'format' => [
            'type' => FieldMapperInterface::MAPPING_CONSTANT,
            'mapping' => 'full_html',
          ],
        ],
        'a_plain_text' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'rich_text_2',
          ],
          'format' => [
            'type' => FieldMapperInterface::MAPPING_CONSTANT,
            'mapping' => 'plain_text',
          ],
        ],
        'a_boolean' => [
          'value' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'status',
          ],
        ],
        'ref' => [
          'target_id' => [
            'type' => FieldMapperInterface::MAPPING_SIMPLE,
            'mapping' => 'refs/*',
          ],
        ],
      ],
    ]);
    $type->save();

    // Create the user with all needed permissions.
    $this->account = $this->drupalCreateUser([
      'administer external entity types',
      'view simple_external_entity external entity',
      'view simple_external_entity external entity collection',
      'access site reports',
    ]);
    $this->drupalLogin($this->account);
  }

  /**
   * Tests creation of a rule and then triggering its execution.
   */
  public function testSimpleExternalEntity() {
    /** @var \Drupal\Tests\WebAssert $assert */
    $assert = $this->assertSession();

    $this->drupalGet('admin/structure/external-entity-types');
    $assert->pageTextContains('Simple external entity');

    // Check "ref test endpoint".
    $ref_json = Json::decode($this->drupalGet('external-entities-test/ref'));
    $this
      ->assertSession()
      ->statusCodeEquals(200);
    $this
      ->assertCount(4, $ref_json);
    $this
      ->assertSession()
      ->responseHeaderEquals('Content-Type', 'application/json');

    // Check "simple test endpoint".
    $simple_json = Json::decode($this->drupalGet('external-entities-test/simple'));
    $this
      ->assertSession()
      ->statusCodeEquals(200);
    $this
      ->assertCount(2, $simple_json);
    $this
      ->assertSession()
      ->responseHeaderEquals('Content-Type', 'application/json');

    // Entity list check.
    $this->drupalGet('simple-external-entity');
    $assert->pageTextContainsOnce('Simple title 1');
    $assert->pageTextContainsOnce('Simple title 2');

    // Entity 1 test.
    $this->drupalGet('simple-external-entity/2596b1ba-43bb-4440-9f0c-f1974f733336');
    $assert->pageTextContains('Simple title 1');
    $assert->pageTextContainsOnce('Just a short string');
    $assert->pageTextContainsOnce('A fixed string');
    $assert->pageTextNotContains('Term 1');
    $assert->pageTextContainsOnce('Term 2');
    $assert->pageTextContainsOnce('Term 3');
    $assert->responseContains('<div>On</div>');
    $assert->pageTextContainsOnce('Some HTML tags');
    $assert->pageTextNotContains('<h2>Some HTML tags</h2>');
    $assert->pageTextContainsOnce('<h2>Other HTML tags</h2>');

    $this->drupalGet('simple-external-entity/2596b1ba-43bb-4440-9f0c-f1974f733337');
    $assert->pageTextContains('Simple title 2');
    $assert->pageTextContainsOnce('Just another short string');
    $assert->pageTextContainsOnce('A fixed string');
    $assert->responseContains('<div>Off</div>');

    // Edit test.
    $this->drupalGet('simple-external-entity/2596b1ba-43bb-4440-9f0c-f1974f733336/edit');
    // Change title.
    $this->fillField('Title', 'Updated title 1');
    // Uncheck 'a_boolean'.
    $this->fillField('a_boolean', FALSE);
    // Remove Term 2.
    try {
      // Drupal 10 use a "Remove" button.
      $this->pressButton('Remove');
    }
    catch (ElementNotFoundException $e) {
      // Drupal 9 uses autocomplete field.
      $this->fillField('ref[0][target_id]', '');
    }
    $this->pressButton('Save');
    $assert->pageTextContains('Updated title 1');
    $assert->pageTextContainsOnce('Just a short string');
    $assert->pageTextContainsOnce('A fixed string');
    $assert->pageTextNotContains('Term 1');
    $assert->pageTextNotContains('Term 2');
    $assert->pageTextContainsOnce('Term 3');
    $assert->responseContains('<div>Off</div>');
  }

}
